@@ -1,4 +1,4 @@
-# Copyright 2008-2020 Richard Dymond (rjdymond@gmail.com)
+# Copyright 2008-2021 Richard Dymond (rjdymond@gmail.com)
 #
 # This file is part of SkoolKit.
 #
@@ -33,11 +33,11 @@ def clock(quiet, prefix, operation, *args, **kwargs):
         info('{} ({:0.2f}s)'.format(prefix, stop - go))
     return result
 
-def run(skoolfile, options):
+def run(skoolfile, options, config):
     # Read custom ASM templates
     t_parser = RefParser()
-    if options.templates:
-        t_parser.parse(options.templates, '#')
+    if config['Templates']:
+        t_parser.parse(config['Templates'], '#')
     templates = {t: t_parser.get_section(t, trim=False) for t in TEMPLATES if t_parser.has_section(t)}
 
     # Create the parser
@@ -46,9 +46,10 @@ def run(skoolfile, options):
     else:
         fname = skoolfile
     asm_mode = options.asm_mode + 4 * int(options.force)
+    label_fmt = (config['EntryLabel'], config['EntryPointLabel'])
     parser = clock(options.quiet, 'Parsed {}'.format(fname), SkoolParser, skoolfile,
-                   options.case, options.base, asm_mode, options.warn, options.fix_mode,
-                   False, options.create_labels, True, options.start, options.end, options.variables)
+                   options.case, options.base, asm_mode, options.warn, options.fix_mode, False,
+                   options.create_labels, True, label_fmt, options.start, options.end, options.variables)
 
     # Write the ASM file
     cls_name = options.writer or parser.asm_writer_class
@@ -65,7 +66,7 @@ def run(skoolfile, options):
             properties[name] = value
     if not options.warn:
         properties['warnings'] = '0'
-    asm_writer = asm_writer_class(parser, properties, templates)
+    asm_writer = asm_writer_class(parser, properties, templates, config)
     clock(options.quiet, 'Wrote ASM to stdout', asm_writer.write)
 
 def main(args):
@@ -118,7 +119,7 @@ def main(args):
     group.add_argument('-u', '--upper', dest='case', action='store_const', const=CASE_UPPER, default=config['Case'],
                        help="Write the disassembly in upper case.")
     group.add_argument('--var', dest='variables', metavar='name=value', type=variable, action='append', default=[],
-                       help="Define a variable that can be used by @if and the SMPL\nmacros. This option may be used multiple times.")
+                       help="Define a variable that can be used by @if and skool\nmacros. This option may be used multiple times.")
     group.add_argument('-V', '--version', action='version', version='SkoolKit {}'.format(VERSION),
                        help='Show SkoolKit version number and exit.')
     group.add_argument('-w', '--no-warnings', dest='warn', action='store_const', const=0, default=config['Warnings'],
@@ -139,4 +140,4 @@ def main(args):
         namespace.asm_mode = 3
     elif namespace.asm_mode == 3:
         namespace.fix_mode = max(namespace.fix_mode, 1)
-    run(namespace.skoolfile, namespace)
+    run(namespace.skoolfile, namespace, config)
