@@ -1,4 +1,4 @@
-# Copyright 2008-2020 Richard Dymond (rjdymond@gmail.com)
+# Copyright 2008-2022 Richard Dymond (rjdymond@gmail.com)
 #
 # This file is part of SkoolKit.
 #
@@ -31,17 +31,17 @@ TEMPLATES = {
     'instruction': '{indent}{operation:{width}} {sep} {text}',
     'label': '{label}{suffix}',
     'org': '{indent}{org} {address}',
-    'register': '; {prefix:>{prefix_len}}{reg:{reg_len}} {text}'
+    'register': '; {prefix:>{prefix_len}}{reg:<{reg_len}} {text}'
 }
 
 class AsmWriter:
-    def __init__(self, parser, properties, templates):
+    def __init__(self, parser, properties, templates, config):
         self.parser = parser
         self.templates = TEMPLATES.copy()
         self.templates.update(templates)
         self.show_warnings = self._get_int_property(properties, 'warnings', 1)
-
-        self.fields = parser.fields
+        self.asm_address_template = config['Address']
+        self.sep_blocks = True
 
         # Build a label dictionary
         self.labels = {}
@@ -97,7 +97,7 @@ class AsmWriter:
 
         self.list_parser = ListParser(properties.get('bullet', '*'))
 
-        self.to_chr = lambda n: chr(n)
+        self.to_chr = chr
         self.get_reg = lambda r: r
         self.space = ' '
         self.pc = 0
@@ -125,6 +125,11 @@ class AsmWriter:
     def case(self):
         return self.parser.case
 
+    # API
+    @property
+    def fields(self):
+        return self.parser.fields
+
     def _get_int_property(self, properties, name, default):
         try:
             return int(properties[name])
@@ -222,9 +227,19 @@ class AsmWriter:
         """
         self._snapshots.append((self.snapshot[:], name))
 
+    def expand_audio(self, text, index):
+        if self.handle_unsupported_macros:
+            return skoolmacro.parse_audio(self, text, index)[0], ''
+        raise skoolmacro.UnsupportedMacroError()
+
+    def expand_copy(self, text, index):
+        if self.handle_unsupported_macros:
+            return skoolmacro.parse_copy(text, index, self.fields)
+        raise skoolmacro.UnsupportedMacroError()
+
     def expand_font(self, text, index):
         if self.handle_unsupported_macros:
-            return skoolmacro.parse_font(text, index)[0], ''
+            return skoolmacro.parse_font(text, index, self.fields)[0], ''
         raise skoolmacro.UnsupportedMacroError()
 
     def expand_html(self, text, index):
@@ -232,7 +247,7 @@ class AsmWriter:
         return end, ''
 
     def expand_include(self, text, index):
-        end, paragraphs, section = skoolmacro.parse_include(text, index)
+        end, paragraphs, section = skoolmacro.parse_include(text, index, self.fields)
         return end, ''
 
     def expand_link(self, text, index):
@@ -244,8 +259,18 @@ class AsmWriter:
     def expand_list(self, text, index):
         return self._ignore_block(text, index, LIST_MARKER, LIST_END_MARKER)
 
+    def expand_over(self, text, index):
+        if self.handle_unsupported_macros:
+            return skoolmacro.parse_over(text, index, self.fields)
+        raise skoolmacro.UnsupportedMacroError()
+
+    def expand_plot(self, text, index):
+        if self.handle_unsupported_macros:
+            return skoolmacro.parse_plot(text, index, self.fields)
+        raise skoolmacro.UnsupportedMacroError()
+
     def expand_r(self, text, index):
-        end, addr_str, address, code_id, anchor, link_text = skoolmacro.parse_r(text, index)
+        end, addr_str, address, code_id, anchor, link_text = skoolmacro.parse_r(self.fields, text, index)
         if link_text:
             return end, link_text
         if code_id:
@@ -254,12 +279,15 @@ class AsmWriter:
         if label is None:
             if self.base_address <= address <= self.end_address:
                 self.warn('Could not convert address {} to label'.format(addr_str))
-            label = self.parser.get_instruction_addr_str(address, addr_str)
+            if self.asm_address_template:
+                label = self.asm_address_template.format(address=address)
+            else:
+                label = self.parser.get_instruction_addr_str(address, addr_str)
         return end, label
 
     def expand_scr(self, text, index):
         if self.handle_unsupported_macros:
-            return skoolmacro.parse_scr(text, index)[0], ''
+            return skoolmacro.parse_scr(text, index, self.fields)[0], ''
         raise skoolmacro.UnsupportedMacroError()
 
     def expand_table(self, text, index):
@@ -267,18 +295,23 @@ class AsmWriter:
 
     def expand_udg(self, text, index):
         if self.handle_unsupported_macros:
-            return skoolmacro.parse_udg(text, index)[0], ''
+            return skoolmacro.parse_udg(text, index, self.fields)[0], ''
         raise skoolmacro.UnsupportedMacroError()
 
     def expand_udgarray(self, text, index):
         if self.handle_unsupported_macros:
             if index < len(text) and text[index] == '*':
-                end = skoolmacro.parse_udgarray_with_frames(text, index)[0]
+                end = skoolmacro.parse_udgarray_with_frames(text, index, self.fields)[0]
             else:
-                end = skoolmacro.parse_udgarray(text, index)[0]
+                end = skoolmacro.parse_udgarray(text, index, fields=self.fields)[0]
             return end, ''
         raise skoolmacro.UnsupportedMacroError()
 
+    def expand_udgs(self, text, index):
+        if self.handle_unsupported_macros:
+            return skoolmacro.parse_udgs(self, text, index)
+        raise skoolmacro.UnsupportedMacroError()
+
     def expand_udgtable(self, text, index):
         return self._ignore_block(text, index, UDGTABLE_MARKER, TABLE_END_MARKER, '')
 
@@ -293,7 +326,10 @@ class AsmWriter:
         except ValueError:
             raise SkoolParsingError("Missing end marker: {}...".format(text[index - len(marker):index + 15]))
         if rep is None:
-            rep = BLOCK_SEP + text[index - len(marker):end] + BLOCK_SEP
+            if self.sep_blocks:
+                rep = BLOCK_SEP + text[index - len(marker):end] + BLOCK_SEP
+            else:
+                rep = marker + self.expand(text[index:end])
         return -end, rep
 
     def format(self, text, width):
