@@ -1,4 +1,4 @@
-# Copyright 2009-2015, 2017-2020 Richard Dymond (rjdymond@gmail.com)
+# Copyright 2009-2015, 2017-2022 Richard Dymond (rjdymond@gmail.com)
 #
 # This file is part of SkoolKit.
 #
@@ -16,6 +16,7 @@
 
 import argparse
 import glob
+import os.path
 
 from skoolkit import info, integer, VERSION
 from skoolkit.config import get_config, show_config, update_options
@@ -23,15 +24,25 @@ from skoolkit.ctlparser import CtlParser
 from skoolkit.snapshot import make_snapshot
 from skoolkit.snaskool import SkoolWriter
 
-def get_ctl_parser(ctlfiles, infile, start, end, def_start, def_end):
+def get_ctl_parser(ctls, infile, start, end, def_start, def_end, defb=None, config=None):
     if infile[-4:].lower() in ('.bin', '.sna', '.szx', '.z80'):
         prefix = infile[:-4]
     else:
         prefix = infile
-    if not ctlfiles:
-        ctlfiles.extend(sorted(glob.glob(prefix + '*.ctl')))
-    if ctlfiles and '0' not in ctlfiles:
+    if not ctls:
+        ctls.extend(sorted(glob.glob(prefix + '*.ctl')))
+    ctlfiles = []
+    if ctls and '0' not in ctls:
         # Use control file(s)
+        for ctl in ctls:
+            if os.path.isdir(ctl):
+                ctlfiles.extend(sorted(glob.glob(os.path.join(ctl, '*.ctl'))))
+            else:
+                ctlfiles.append(ctl)
+    if defb:
+        ctl_parser = CtlParser({def_start: 'b', def_end: 'i'})
+        config['DefbSize'] = defb
+    elif ctlfiles:
         if len(ctlfiles) > 1:
             suffix = 's'
         else:
@@ -45,9 +56,11 @@ def get_ctl_parser(ctlfiles, infile, start, end, def_start, def_end):
 
 def run(infile, options, config):
     snapshot, start, end = make_snapshot(infile, options.org, options.start, options.end, options.page)
-    ctl_parser = get_ctl_parser(options.ctlfiles, infile, options.start, options.end, start, end)
+    if options.start is None:
+        options.start = 0
+    ctl_parser = get_ctl_parser(options.ctls, infile, options.start, options.end, start, end, options.defb, config)
     writer = SkoolWriter(snapshot, ctl_parser, options, config)
-    writer.write_skool(config['ListRefs'], config['Text'])
+    writer.write_skool()
 
 def main(args):
     config = get_config('sna2skool')
@@ -59,9 +72,12 @@ def main(args):
     )
     parser.add_argument('snafile', help=argparse.SUPPRESS, nargs='?')
     group = parser.add_argument_group('Options')
-    group.add_argument('-c', '--ctl', dest='ctlfiles', metavar='FILE', action='append', default=[],
-                       help="Use FILE as a control file. FILE may be '-' for standard input, or '0' to use no control file. "
+    group.add_argument('-c', '--ctl', dest='ctls', metavar='PATH', action='append', default=[],
+                       help="Specify a control file to use, or a directory from which to read control files. "
+                            "PATH may be '-' for standard input, or '0' to use no control file. "
                             "This option may be used multiple times.")
+    group.add_argument('-d', '--defb', dest='defb', metavar='SIZE', type=int,
+                       help='Disassemble as DEFB statements of this size.')
     group.add_argument('-e', '--end', dest='end', metavar='ADDR', type=integer, default=65536,
                        help='Stop disassembling at this address (default=65536).')
     group.add_argument('-H', '--hex', dest='base', action='store_const', const=16, default=config['Base'],
@@ -76,8 +92,8 @@ def main(args):
                        help='Specify the page (0-7) of a 128K snapshot to map to 49152-65535.')
     group.add_argument('--show-config', dest='show_config', action='store_true',
                        help="Show configuration parameter values.")
-    group.add_argument('-s', '--start', dest='start', metavar='ADDR', type=integer, default=0,
-                       help='Start disassembling at this address (default=16384).')
+    group.add_argument('-s', '--start', dest='start', metavar='ADDR', type=integer,
+                       help='Start disassembling at this address.')
     group.add_argument('-V', '--version', action='version', version='SkoolKit {}'.format(VERSION),
                        help='Show SkoolKit version number and exit.')
     group.add_argument('-w', '--line-width', dest='line_width', metavar='W', type=int, default=config['LineWidth'],
