@@ -1,4 +1,4 @@
-# Copyright 2009-2020 Richard Dymond (rjdymond@gmail.com)
+# Copyright 2009-2022 Richard Dymond (rjdymond@gmail.com)
 #
 # This file is part of SkoolKit.
 #
@@ -15,6 +15,7 @@
 # SkoolKit. If not, see <http://www.gnu.org/licenses/>.
 
 from collections import defaultdict, namedtuple
+import re
 
 from skoolkit import (SkoolKitError, warn, write_line, wrap, parse_int,
                       get_address_format, format_template)
@@ -22,27 +23,30 @@ from skoolkit.components import get_component, get_value
 from skoolkit.skoolasm import UDGTABLE_MARKER
 from skoolkit.skoolctl import (AD_LABEL, AD_REFS, TITLE, DESCRIPTION,
                                REGISTERS, MID_BLOCK, INSTRUCTION, END)
-from skoolkit.skoolmacro import ClosingBracketError, parse_brackets
+from skoolkit.skoolmacro import INTEGER, ClosingBracketError, parse_brackets
 from skoolkit.skoolparser import (get_address, parse_asm_refs_directive,
                                   parse_register, LIST_MARKER, TABLE_MARKER,
                                   LIST_END_MARKER, TABLE_END_MARKER)
+from skoolkit.z80 import get_timing
 
 MIN_COMMENT_WIDTH = 10
 AD_LABEL_PREFIX = AD_LABEL + '='
 AD_REFS_PREFIX = AD_REFS + '='
 
-DisassemblerConfig = namedtuple('DisassemblerConfig', 'asm_hex asm_lower defb_size defm_size defw_size')
+DisassemblerConfig = namedtuple('DisassemblerConfig', 'asm_hex asm_lower defb_size defm_size defw_size wrap')
 
+# Component API
 def calculate_references(entries, operations):
     """
-    For each entry point in each routine, calculate a list of the entries
-    containing instructions that jump to or call that entry point.
+    For each instruction address in a memory map entry, calculate a list of the
+    entries containing instructions that jump to, call or otherwise refer to
+    that address.
 
     :param entries: A collection of memory map entries.
-    :param operations: A tuple of operation prefixes. Any instruction whose
-                       operation starts with one of these prefixes is regarded
-                       as a jump or call operation, and therefore identifies an
-                       entry point.
+    :param operations: A tuple of regular expression patterns. The address
+                       operand of any instruction whose operation matches one
+                       of these patterns identifies an entry point that will be
+                       marked with an asterisk in the skool file.
     :return: A dictionary of entry point addresses.
     """
     instructions = {i.address: (i, e) for e in entries for i in e.instructions}
@@ -50,7 +54,7 @@ def calculate_references(entries, operations):
     for entry in entries:
         for instruction in entry.instructions:
             operation = instruction.operation
-            if operation.upper().startswith(operations):
+            if any(re.match(op, operation.upper()) for op in operations):
                 addr_str = get_address(operation)
                 if addr_str:
                     ref_addr = parse_int(addr_str)
@@ -107,7 +111,7 @@ class Entry:
         self.bad_blocks = []
         for sub_block in sub_blocks:
             last_instruction = sub_block.instructions[-1]
-            if last_instruction.address + len(last_instruction.bytes) > sub_block.end:
+            if last_instruction.address + len(last_instruction.bytes) > sub_block.end < 65536:
                 self.bad_blocks.append(sub_block)
 
     def width(self):
@@ -124,8 +128,9 @@ class Disassembly:
             asm_hex,
             asm_lower,
             self.config.get('DefbSize', 8),
-            self.config.get('DefmSize', 66),
-            self.config.get('DefwSize', 1)
+            self.config.get('DefmSize', 65),
+            self.config.get('DefwSize', 1),
+            self.config.get('Wrap', 0)
         )
         self.disassembler = get_component('Disassembler', snapshot, dconfig)
         self.ref_calc = get_component('SnapshotReferenceCalculator')
@@ -235,8 +240,12 @@ class Disassembly:
                     instruction.refs, instruction.rrefs = parse_asm_refs_directive(asm_dir)
 
     def _calculate_references(self, self_refs):
-        operations = get_value('SnapshotReferenceOperations').upper()
-        referrers = self.ref_calc.calculate_references(self.entries, tuple(operations.split(',')))
+        regexes = get_value('SnapshotReferenceOperations')
+        if regexes and not regexes[0].isalpha():
+            operations = regexes[1:].split(regexes[0])
+        else:
+            operations = regexes.split(',')
+        referrers = self.ref_calc.calculate_references(self.entries, tuple(p.replace('\\i', INTEGER) for p in operations))
         for entry in self.entries:
             for instruction in entry.instructions:
                 for ref_entry in referrers.get(instruction.address, ()):
@@ -252,6 +261,7 @@ class SkoolWriter:
         self.disassembly = Disassembly(snapshot, ctl_parser, config, self.asm_hex, options.case == 1)
         self.address_fmt = get_address_format(self.asm_hex, options.case == 1)
         self.config = config
+        self.list_refs = config['ListRefs']
 
     def address_str(self, address, pad=True):
         if self.asm_hex or pad:
@@ -265,13 +275,13 @@ class SkoolWriter:
             lines.pop()
         return lines
 
-    def write_skool(self, write_refs, text):
+    def write_skool(self):
         for entry_index, entry in enumerate(self.disassembly.entries):
             if entry_index:
                 write_line('')
-            self._write_entry(entry, write_refs, text)
+            self._write_entry(entry)
 
-    def _write_entry(self, entry, write_refs, show_text):
+    def _write_entry(self, entry):
         if entry.header:
             for line in entry.header:
                 write_line(line)
@@ -290,12 +300,12 @@ class SkoolWriter:
 
         if entry.has_title:
             self.write_comment(entry.title)
-            wrote_desc = self._write_entry_description(entry, write_refs)
+            wrote_desc = self._write_entry_description(entry)
             wrote_desc = self._write_registers(entry, wrote_desc)
         else:
             wrote_desc = False
 
-        self._write_body(entry, wrote_desc, write_refs, show_text and entry.ctl != 't')
+        self._write_body(entry, wrote_desc)
 
         self.write_asm_directives(entry.get_ignoreua_directive(END))
         self.write_paragraphs(entry.end_comment)
@@ -305,12 +315,12 @@ class SkoolWriter:
             for line in entry.footer:
                 write_line(line)
 
-    def _write_entry_description(self, entry, write_refs):
+    def _write_entry_description(self, entry):
         wrote_desc = False
         ignoreua_d = entry.get_ignoreua_directive(DESCRIPTION)
-        if write_refs:
+        if self.list_refs:
             referrers = entry.instructions[0].referrers
-            if referrers and (write_refs == 2 or not entry.description):
+            if referrers and (self.list_refs == 2 or not entry.description):
                 self.write_comment('')
                 self.write_asm_directives(ignoreua_d)
                 self.write_referrers(referrers, False)
@@ -362,24 +372,40 @@ class SkoolWriter:
         for instruction in block.instructions:
             instruction.comment = [None]
             if block.comment:
-                instruction.comment[0] = block.comment.pop(0)[1]
-                while block.comment and block.comment[0][0]:
-                    instruction.comment.append(block.comment.pop(0)[1])
+                if block.repeat_comment:
+                    instruction.comment[:] = [c[1] for c in block.comment]
+                else:
+                    instruction.comment[0] = block.comment.pop(0)[1]
+                    while block.comment and block.comment[0][0]:
+                        instruction.comment.append(block.comment.pop(0)[1])
             elif show_text:
                 instruction.comment[0] = self.to_ascii(instruction.bytes)
+            elif self.config['Timings']:
+                tstates = get_timing(instruction)
+                if isinstance(tstates, tuple):
+                    instruction.comment[0] = '[{}/{}]'.format(*tstates)
+                elif tstates:
+                    instruction.comment[0] = f'[{tstates}]'
             elif closing:
                 instruction.comment[0] = ''
-        final_comment = block.instructions[-1].comment
-        final_comment.extend(t[1] for t in block.comment)
+
+        if block.comment and block.repeat_comment:
+            instructions = block.instructions
+        else:
+            final_comment = block.instructions[-1].comment
+            final_comment.extend(t[1] for t in block.comment)
+            instructions = block.instructions[-1:]
         if closing:
-            if len(final_comment[-1]) + len(closing) <= width:
-                final_comment[-1] = (final_comment[-1] + closing).lstrip()
-            else:
-                final_comment.append(closing.lstrip())
+            for instruction in instructions:
+                comment = instruction.comment
+                if len(comment[-1]) + len(closing) <= width:
+                    comment[-1] = (comment[-1] + closing).lstrip()
+                else:
+                    comment.append(closing.lstrip())
 
     def _format_instruction_comments(self, block, width, show_text):
         comment = ''.join(t[1] for t in block.comment)
-        multi_line = len(block.instructions) > 1 and (comment or len(block.comment) > 1)
+        multi_line = len(block.instructions) > 1 and (comment or len(block.comment) > 1) and not block.repeat_comment
         opening = closing = ''
         if multi_line and len(block.comment) == 1 and not comment.replace('.', ''):
             comment = comment[1:]
@@ -403,15 +429,16 @@ class SkoolWriter:
             block.comment[0] = (0, opening + block.comment[0][1])
         self._set_instruction_comments(block, width, closing, show_text)
 
-    def _write_body(self, entry, wrote_desc, write_refs, show_text):
+    def _write_body(self, entry, wrote_desc):
+        show_text = self.config['Text'] and entry.ctl != 't'
         op_width = max((self.config['InstructionWidth'], entry.width()))
         comment_width = max(self.comment_width - op_width - 8, self.config['CommentWidthMin'])
         for index, block in enumerate(entry.blocks):
             ignoreua_m = block.get_ignoreua_directive(MID_BLOCK, block.start)
             begun_header = False
-            if index > 0 and entry.ctl == 'c' and write_refs:
+            if index > 0 and entry.ctl == 'c' and self.list_refs:
                 referrers = block.instructions[0].referrers
-                if referrers and (write_refs == 2 or not block.header):
+                if referrers and (self.list_refs == 2 or not block.header):
                     self.write_asm_directives(ignoreua_m)
                     self.write_referrers(referrers)
                     begun_header = True
@@ -428,15 +455,15 @@ class SkoolWriter:
                     self.write_asm_directives(ignoreua_m)
                 self.write_paragraphs(block.header)
             self._format_instruction_comments(block, comment_width, show_text)
-            self._write_instructions(entry, block, op_width, write_refs)
+            self._write_instructions(entry, block, op_width)
 
-    def _write_instructions(self, entry, block, op_width, write_refs):
+    def _write_instructions(self, entry, block, op_width):
         for index, instruction in enumerate(block.instructions):
             ctl = instruction.ctl or ' '
             address = instruction.address
             operation = instruction.operation
             comment = instruction.comment.pop(0)
-            if index > 0 and entry.ctl == 'c' and ctl == '*' and write_refs:
+            if index > 0 and entry.ctl == 'c' and ctl == '*' and self.list_refs:
                 self.write_referrers(instruction.referrers)
             self.write_asm_directives(*instruction.asm_directives)
             self.write_asm_directives(block.get_ignoreua_directive(INSTRUCTION, instruction.address))
@@ -477,14 +504,15 @@ class SkoolWriter:
 
     def write_referrers(self, referrers, erefs=True):
         if referrers:
+            ref_fmt = self.config['RefFormat']
             if erefs:
                 key = 'EntryPointRef'
             else:
                 key = 'Ref'
-            fields = {'ref': '#R' + self.address_str(referrers[-1], False)}
+            fields = {'ref': ref_fmt.format(address=self.address_str(referrers[-1], False))}
             if len(referrers) > 1:
                 key += 's'
-                fields['refs'] = ', '.join(['#R' + self.address_str(r, False) for r in referrers[:-1]])
+                fields['refs'] = ', '.join([ref_fmt.format(address=self.address_str(r, False)) for r in referrers[:-1]])
             self.write_comment([format_template(self.config[key], key, **fields)])
 
     def write_asm_directives(self, *directives):
