@@ -24,9 +24,10 @@ from urllib.parse import urlparse
 
 from skoolkit import (SkoolKitError, get_dword, get_int_param, get_object,
                       get_word, get_word3, integer, open_file, read_bin_file,
-                      warn, write, write_line, ROM48, VERSION)
-from skoolkit.basic import TextReader
-from skoolkit.simulator import Simulator
+                      warn, write_line, ROM48, VERSION)
+from skoolkit.loadtracer import LoadTracer
+from skoolkit.simulator import (Simulator, A, F, B, C, D, E, H, L, IXh, IXl, IYh, IYl,
+                                SP, I, R, xA, xF, xB, xC, xD, xE, xH, xL, PC)
 from skoolkit.snapshot import move, poke, print_reg_help, print_state_help, write_z80v3
 
 SYSVARS = (
@@ -133,14 +134,17 @@ SIM_LOAD_PATCH = {
     0x5C04: 0x050D, # KSTATE4 0/1
     0x5C06: 0x0D23, # KSTATE4 2/3
     0x5C08: 0x0D,   # LAST-K
+    0x5C3A: 0xFF,   # ERR-NR
     0x5C3B: 0x0C,   # FLAGS
     0x5C3C: 0x01,   # TV-FLAG
-    0x5C3D: 0x50,   # ERR-SP
+    0x5C47: 0x01,   # SUBPPC
     0x5C5B: 0x5CCF, # K-CUR
+    0x5C5D: 0x5CCD, # CH-ADD
     0x5C61: 0x5CD1, # WORKSP
     0x5C63: 0x5CD1, # STKBOT
     0x5C65: 0x5CD1, # STKEND
     0x5C68: 0x5C92, # MEM
+    0x5C74: 0x1AE1, # T-ADDR
     0x5C82: 0x19,   # ECHO-E
     0x5C86: 0xE0,   # DF-CCL
     0x5C8A: 0x21,   # S-POSNL
@@ -149,9 +153,9 @@ SIM_LOAD_PATCH = {
     0x5CCD: 0x2222, # ""
     0x5CCF: 0x0D,   # ENTER
     0x5CD0: 0x80,   # End of program area
-    0xFF50: 0x107F, # Address of ED-ERROR
-    0xFF52: 0xFF54, # Address of next item on the stack
-    0xFF54: 0x12B4, # Address of entry point in 'MAIN EXECUTION' loop at 0x12A2
+    0xFF50: 0x1B52, # Address of SCAN-LOOP
+    0xFF52: 0x1B76, # Address of STMT-RET
+    0xFF54: 0x12B7, # Address of entry point in 'MAIN EXECUTION' loop at 0x12A2
     0xFF56: 0x3E00  # RAMTOP marker
 }
 
@@ -166,14 +170,6 @@ SIM_LOAD_CODE_PATCH = {
     0x5CD1: 0x80    # End of program area
 }
 
-SILENCE = 0
-PILOT = 1
-SYNC = 2
-DATA = 3
-STOP = 4
-
-SIM_TIMEOUT = 10 * 60 * 3500000 # 10 minutes of Z80 CPU time
-
 class SkoolKitArgumentParser(argparse.ArgumentParser):
     def convert_arg_line_to_args(self, arg_line):
         for arg in arg_line.split():
@@ -197,208 +193,6 @@ TAP_TIMINGS_HEADER = TapeBlockTimings(8063, 2168, (667, 735), 855, 1710, 3500000
 
 TAP_TIMINGS_DATA = TapeBlockTimings(3223, 2168, (667, 735), 855, 1710, 3500000)
 
-def get_ear_samples(blocks):
-    samples = []
-    tstates = -1
-    sample = 0xFF # EAR off (bit 6 set), no keys pressed
-    ear = 0x40
-
-    for i, (timings, data) in enumerate(blocks):
-        # Pilot tone
-        for n in range(timings.pilot_len):
-            sample ^= ear
-            if tstates < 0:
-                tstates = 0
-            else:
-                tstates += timings.pilot
-            samples.append((tstates, sample, PILOT, i))
-
-        # Sync pulses
-        for s in timings.sync:
-            sample ^= ear
-            tstates += s
-            samples.append((tstates, sample, SYNC, i))
-
-        # Data
-        for b in data:
-            for j in range(8):
-                if b & 0x80:
-                    duration = timings.one
-                else:
-                    duration = timings.zero
-                for k in range(2):
-                    sample ^= ear
-                    tstates += duration
-                    samples.append((tstates, sample, DATA, i))
-                b <<= 1
-
-        # Pause
-        if i + 1 < len(blocks) and timings.pause:
-            r = samples.pop()
-            samples.append((r[0], r[1], SILENCE, i + 1))
-            tstates += timings.pause
-
-    ts, sample, stype, bnum = samples[-1]
-    samples.append((ts + 3500, sample, STOP, bnum))
-
-    return samples
-
-class LoadTracer:
-    def __init__(self, blocks, stop):
-        self.blocks = [b[1] for b in blocks]
-        self.samples = get_ear_samples(blocks)
-        self.index = 0
-        self.max_index = len(self.samples) - 1
-        self.stop = stop
-        self.tape_started = None
-        self.progress = 0
-        self.tape_length = self.samples[-1][0] // 1000
-        self.end_of_tape = 0
-        self.tape_end_time = 0
-        self.pulse_type = None
-        self.custom_loader = False
-        self.border = 7
-        self.text = TextReader()
-
-    def trace(self, simulator, instruction):
-        if self.tape_started is not None:
-            offset = simulator.tstates - self.tape_started
-            block_num = self.samples[self.index][3]
-            while self.index < self.max_index and self.samples[self.index + 1][0] < offset:
-                self.index += 1 # pragma: no cover
-            if self.samples[self.index][3] > block_num:
-                # Pause tape between blocks
-                self.tape_started = None # pragma: no cover
-            else:
-                if self.index == self.max_index:
-                    self.end_of_tape += 1
-                    if self.end_of_tape == 1:
-                        write_line('Tape finished')
-                        self.tape_end_time = simulator.tstates
-                if self.custom_loader: # pragma: no cover
-                    progress = self.samples[self.index][0] // self.tape_length
-                    if progress > self.progress:
-                        msg = f'[{progress/10:0.1f}%]'
-                        write(msg + chr(8) * len(msg))
-                        self.progress = progress
-
-        if simulator.pc == self.stop:
-            write_line(f'Simulation stopped (PC at start address): PC={simulator.pc}')
-            return True
-
-        if simulator.pc == 0x0556:
-            return self.fast_load(simulator)
-
-        if self.end_of_tape and self.stop is None:
-            if self.custom_loader: # pragma: no cover
-                write_line(f'Simulation stopped (end of tape): PC={simulator.pc}')
-                return True
-            if simulator.pc > 0x3FFF:
-                write_line(f'Simulation stopped (PC in RAM): PC={simulator.pc}')
-                return True
-            if simulator.tstates - self.tape_end_time > 3500000: # pragma: no cover
-                write_line(f'Simulation stopped (tape ended 1 second ago): PC={simulator.pc}')
-                return True
-
-        if simulator.tstates > SIM_TIMEOUT: # pragma: no cover
-            write_line(f'Simulation stopped (timed out): PC={simulator.pc}')
-            return True
-
-    def read_port(self, simulator, port):
-        if port & 0xFF == 0xFE:
-            read_tape = False
-            if simulator.pc > 0x7FFF: # pragma: no cover
-                self.custom_loader = True
-                read_tape = True
-            elif 0x0562 <= simulator.pc <= 0x05F1:
-                read_tape = True # pragma: no cover
-
-            if read_tape: # pragma: no cover
-                if self.tape_started is None:
-                    self.tape_started = simulator.tstates - self.samples[self.index][0]
-
-                if self.index == self.max_index - 1:
-                    ts, sample, stype, bnum = self.samples.pop()
-                    self.samples.append((self.samples[-1][0], sample, STOP, bnum))
-
-                pulse_type = self.samples[self.index][2]
-                if pulse_type != self.pulse_type:
-                    if pulse_type == PILOT:
-                        write_line('Pilot tone')
-                    elif pulse_type == SYNC:
-                        write_line('Sync pulses')
-                    elif pulse_type == DATA:
-                        length = len(self.blocks[self.samples[self.index][3]])
-                        write_line(f'Data ({length} bytes)\n')
-                    self.pulse_type = pulse_type
-
-                return self.samples[self.index][1]
-
-    def write_port(self, simulator, port, value):
-        if port & 0x01 == 0:
-            self.border = value & 0x07
-
-    def fast_load(self, simulator):
-        block = self.blocks[self.samples[self.index][3]]
-        registers = simulator.registers
-        ix = registers['IXl'] + 256 * registers['IXh'] # Start address
-        de = registers['E'] + 256 * registers['D'] # Block length
-        a = registers['A']
-        data_len = len(block) - 2
-
-        # Preload the machine stack with 0x053F (as done at 0x055E)
-        registers['H'], registers['L'] = 0x05, 0x3F # SA-LD-RET
-        simulator.push('HL')
-
-        if a == block[0]:
-            skipped = ''
-        else:
-            skipped = ' [skipped]\n'
-        if block[0] == 0 and data_len >= 17 and block[1] <= 3:
-            name = ''.join(self.text.get_chars(b) for b in block[2:12])
-            if block[1] == 3:
-                write_line(f'Bytes: {name}{skipped}')
-            elif block[1] == 2:
-                write_line(f'Character array: {name}{skipped}')
-            elif block[1] == 1:
-                write_line(f'Number array: {name}{skipped}')
-            elif block[1] == 0:
-                write_line(f'Program: {name}{skipped}')
-        elif skipped:
-            write_line(f'Data block ({data_len} bytes){skipped}')
-        else:
-            write_line(f'Fast loading data block: {ix},{de}\n')
-
-        if skipped:
-            registers['F'] = 0x00 # Reset carry flag: error
-        else:
-            if de <= data_len:
-                simulator.snapshot[ix:ix + de] = block[1:1 + de]
-                registers['F'] = 0x01 # Set carry flag: success
-                ix += de
-                de = 0
-            else:
-                simulator.snapshot[ix:ix + data_len + 1] = block[1:]
-                registers['F'] = 0x00 # Reset carry flag: error
-                ix += data_len + 1
-                de -= data_len + 1
-            registers['IXh'] = (ix >> 8) & 0xFF
-            registers['IXl'] = ix & 0xFF
-            registers['D'] = (de >> 8) & 0xFF
-            registers['E'] = de & 0xFF
-
-        simulator.pc = 0x05E2
-
-        block_num = self.samples[self.index][3]
-        while self.index < self.max_index and self.samples[self.index][3] == block_num:
-            self.index += 1
-        if self.samples[self.index][3] > block_num:
-            # Pause tape between blocks
-            self.tape_started = None
-        else:
-            self.tape_started = simulator.tstates - self.samples[self.index][0]
-        self.pulse_type = DATA
-
 def _write_z80(ram, options, fname):
     parent_dir = os.path.dirname(fname)
     if parent_dir and not os.path.isdir(parent_dir):
@@ -406,6 +200,24 @@ def _write_z80(ram, options, fname):
     write_line('Writing {0}'.format(fname))
     write_z80v3(fname, ram, options.reg, options.state)
 
+def _ram_operations(snapshot, ram_ops, blocks=None):
+    counters = {}
+    for spec in ram_ops:
+        op_type, sep, param_str = spec.partition('=')
+        if op_type == 'call':
+            _call(snapshot, param_str)
+        elif op_type == 'load':
+            if blocks:
+                _load(snapshot, counters, blocks, param_str)
+        elif op_type == 'move':
+            move(snapshot, param_str)
+        elif op_type == 'poke':
+            poke(snapshot, param_str)
+        elif op_type == 'sysvars':
+            snapshot[23552:23755] = SYSVARS
+        else:
+            raise SkoolKitError(f'Invalid operation: {spec}')
+
 def sim_load(blocks, options):
     snapshot = [0] * 65536
     rom = read_bin_file(ROM48, 16384)
@@ -422,22 +234,44 @@ def sim_load(blocks, options):
             if b > 0xFF:
                 snapshot[a + 1] = b // 256
     snapshot[0xFF58:] = snapshot[0x3E08:0x3EB0] # UDGs
-    simulator = Simulator(snapshot, {'A': 0x0D, 'SP': 0xFF50})
-    tracer = LoadTracer(blocks, options.start)
-    simulator.add_tracer(tracer)
+    config = {'fast_djnz': True, 'fast_ldir': True}
+    simulator = Simulator(snapshot, {'SP': 0xFF50}, config=config)
+    tracer = LoadTracer(blocks)
+    simulator.set_tracer(tracer)
     try:
-        simulator.run(0x0F3B) # Entry point in EDITOR at 0F2C
+        tracer.run(simulator, 0x0605, options.start) # SAVE-ETC
+        _ram_operations(snapshot, options.ram_ops)
     except KeyboardInterrupt: # pragma: no cover
-        write_line(f'Simulation stopped (interrupted): PC={simulator.pc}')
-    registers = simulator.registers
-    registers['IX'] = registers['IXl'] + 256 * registers['IXh']
-    registers['IY'] = registers['IYl'] + 256 * registers['IYh']
-    registers['PC'] = simulator.pc
-    del registers['IXl'], registers['IXh'], registers['IYl'], registers['IYh']
+        write_line(f'Simulation stopped (interrupted): PC={simulator.registers[PC]}')
+    sim_registers = simulator.registers
+    registers = {
+        'A': sim_registers[A],
+        'F': sim_registers[F],
+        'B': sim_registers[B],
+        'C': sim_registers[C],
+        'D': sim_registers[D],
+        'E': sim_registers[E],
+        'H': sim_registers[H],
+        'L': sim_registers[L],
+        'IX': sim_registers[IXl] + 256 * sim_registers[IXh],
+        'IY': sim_registers[IYl] + 256 * sim_registers[IYh],
+        'I': sim_registers[I],
+        'R': sim_registers[R],
+        'SP': sim_registers[SP],
+        '^A': sim_registers[xA],
+        '^F': sim_registers[xF],
+        '^B': sim_registers[xB],
+        '^C': sim_registers[xC],
+        '^D': sim_registers[xD],
+        '^E': sim_registers[xE],
+        '^H': sim_registers[xH],
+        '^L': sim_registers[xL],
+        'PC': sim_registers[PC]
+    }
     options.reg = [f'{r}={v}' for r, v in registers.items()] + options.reg
     state = [f'im={simulator.imode}', f'iff={simulator.iff2}', f'border={tracer.border}']
     options.state = state + options.state
-    return simulator.snapshot[0x4000:]
+    return simulator.memory[0x4000:]
 
 def _get_load_params(param_str):
     params = []
@@ -503,17 +337,11 @@ def _load(snapshot, counters, blocks, param_str):
 
 def _get_ram(blocks, options):
     snapshot = [0] * 65536
-
-    operations = []
     standard_load = True
     for spec in options.ram_ops:
-        op_type, sep, param_str = spec.partition('=')
-        if op_type in ('call', 'load', 'move', 'poke', 'sysvars'):
-            operations.append((op_type, param_str))
-            if op_type == 'load':
-                standard_load = False
-        else:
-            raise SkoolKitError("Invalid operation: {}".format(spec))
+        if spec.partition('=')[0] == 'load':
+            standard_load = False
+            break
 
     if standard_load:
         start = None
@@ -538,19 +366,7 @@ def _get_ram(blocks, options):
                         _load_block(snapshot, block, start)
                         start = None
 
-    counters = {}
-    for op_type, param_str in operations:
-        if op_type == 'call':
-            _call(snapshot, param_str)
-        elif op_type == 'load':
-            _load(snapshot, counters, blocks, param_str)
-        elif op_type == 'move':
-            move(snapshot, param_str)
-        elif op_type == 'poke':
-            poke(snapshot, param_str)
-        elif op_type == 'sysvars':
-            snapshot[23552:23755] = SYSVARS
-
+    _ram_operations(snapshot, options.ram_ops, blocks)
     return snapshot[16384:]
 
 def _get_tzx_block(data, i, sim):
@@ -563,10 +379,11 @@ def _get_tzx_block(data, i, sim):
         # Standard speed data block
         length = get_word(data, i + 2)
         tape_data = data[i + 4:i + 4 + length]
-        if tape_data[0] == 0:
-            timings = TAP_TIMINGS_HEADER
-        else:
-            timings = TAP_TIMINGS_DATA
+        if tape_data:
+            if tape_data[0] == 0:
+                timings = TAP_TIMINGS_HEADER
+            else:
+                timings = TAP_TIMINGS_DATA
         i += 4 + length
     elif block_id == 17:
         # Turbo speed data block
@@ -693,7 +510,7 @@ def _get_tzx_blocks(data, sim):
 def get_tap_blocks(tap, sim=False):
     blocks = []
     i = 0
-    while i < len(tap):
+    while i + 1 < len(tap):
         block_len = tap[i] + 256 * tap[i + 1]
         i += 2
         data = tap[i:i + block_len]
