@@ -50,11 +50,12 @@ class MockSkoolParser:
         self.case = kwargs.get('case')
         self.create_labels = kwargs.get('create_labels')
         self.asm_labels = kwargs.get('asm_labels')
+        self.label_fmt = kwargs.get('label_fmt')
         self.snapshot = None
         self.entries = {}
         self.memory_map = []
         self.fields = {}
-        self.expands = ()
+        self.expands = []
         mock_skool_parser = self
 
     def get_entry(self, address):
@@ -140,9 +141,9 @@ class Skool2HtmlTest(SkoolKitTestCase):
     @patch.object(skool2html, 'run', mock_run)
     @patch.object(skool2html, 'get_config', mock_config)
     def test_default_option_values(self):
-        infiles = ['game1.ref', 'game2.skool']
+        infiles = ['game.skool', 'data.ref']
         skool2html.main(infiles)
-        files, options = run_args
+        files, options, config = run_args
         self.assertEqual(files, infiles)
         self.assertFalse(options.asm_labels)
         self.assertFalse(options.asm_one_page)
@@ -154,6 +155,7 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertFalse(options.show_timings)
         self.assertEqual(options.config_specs, [])
         self.assertFalse(options.new_images)
+        self.assertFalse(options.new_audio)
         self.assertEqual(options.case, 0)
         self.assertEqual(options.base, 0)
         self.assertEqual(options.files, 'dimoP')
@@ -161,6 +163,8 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertEqual(options.output_dir, '.')
         self.assertEqual(options.params, [])
         self.assertEqual(options.variables, [])
+        self.assertEqual(config['EntryLabel'], 'L{address}')
+        self.assertEqual(config['EntryPointLabel'], '{main}_{index}')
 
     @patch.object(skool2html, 'run', mock_run)
     def test_config_read_from_file(self):
@@ -172,18 +176,21 @@ class Skool2HtmlTest(SkoolKitTestCase):
             Base=16
             Case=-1
             CreateLabels=1
+            EntryLabel=L_{{address}}
+            EntryPointLabel={{main}}__{{index}}
             JoinCss=css.css
             OutputDir={}
             Quiet=1
+            RebuildAudio=1
             RebuildImages=1
             Search=this,that
             Theme=dark,wide
             Time=1
         """.format(output_dir)
         self.write_text_file(dedent(ini).strip(), 'skoolkit.ini')
-        infiles = ['game1.ref', 'game2.skool']
+        infiles = ['game.skool', 'data.ref']
         skool2html.main(infiles)
-        files, options = run_args
+        files, options, config = run_args
         self.assertEqual(files, infiles)
         self.assertTrue(options.asm_labels)
         self.assertTrue(options.asm_one_page)
@@ -194,12 +201,39 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertEqual(options.quiet, 1)
         self.assertTrue(options.show_timings)
         self.assertEqual(options.config_specs, ['Game/AsmSinglePage=1'])
+        self.assertTrue(options.new_audio)
         self.assertTrue(options.new_images)
         self.assertEqual(options.case, -1)
         self.assertEqual(options.base, 16)
         self.assertEqual(options.files, 'dimoP')
         self.assertEqual(options.pages, [])
         self.assertEqual(options.output_dir, output_dir)
+        self.assertEqual(config['EntryLabel'], 'L_{address}')
+        self.assertEqual(config['EntryPointLabel'], '{main}__{index}')
+
+    @patch.object(skool2html, 'get_object', Mock(return_value=TestHtmlWriter))
+    @patch.object(skool2html, 'SkoolParser', MockSkoolParser)
+    @patch.object(skool2html, 'write_disassembly', mock_write_disassembly)
+    def test_config_read_from_file_is_passed(self):
+        output_dir = self.make_directory()
+        ini = """
+            [skool2html]
+            AsmLabels=1
+            Base=16
+            Case=1
+            CreateLabels=1
+            EntryLabel=L_{{address}}
+            EntryPointLabel={{main}}__{{index}}
+        """.format(output_dir)
+        self.write_text_file(dedent(ini).strip(), 'skoolkit.ini')
+        skoolfile = self.write_text_file(suffix='.skool')
+        skool2html.main((skoolfile,))
+        self.assertEqual(mock_skool_parser.skoolfile, skoolfile)
+        self.assertEqual(mock_skool_parser.base, 16)
+        self.assertEqual(mock_skool_parser.case, 1)
+        self.assertTrue(mock_skool_parser.create_labels)
+        self.assertTrue(mock_skool_parser.asm_labels)
+        self.assertEqual(('L_{address}', '{main}__{index}'), mock_skool_parser.label_fmt)
 
     @patch.object(skool2html, 'run', mock_run)
     def test_invalid_option_values_read_from_file(self):
@@ -212,7 +246,7 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.write_text_file(dedent(ini).strip(), 'skoolkit.ini')
         infiles = ['game.skool']
         skool2html.main(infiles)
-        files, options = run_args
+        files, options, config = run_args
         self.assertEqual(files, infiles)
         self.assertFalse(options.asm_labels)
         self.assertFalse(options.asm_one_page)
@@ -229,6 +263,8 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertEqual(options.files, 'dimoP')
         self.assertEqual(options.pages, [])
         self.assertEqual(options.output_dir, '.')
+        self.assertEqual(config['EntryLabel'], 'L{address}')
+        self.assertEqual(config['EntryPointLabel'], '{main}_{index}')
 
     def test_no_arguments(self):
         output, error = self.run_skool2html(catch_exit=2)
@@ -1211,6 +1247,16 @@ class Skool2HtmlTest(SkoolKitTestCase):
             self.assertEqual(error, '')
             self.assertTrue(html_writer.file_info.replace_images)
 
+    @patch.object(skool2html, 'get_object', Mock(return_value=TestHtmlWriter))
+    @patch.object(skool2html, 'SkoolParser', MockSkoolParser)
+    @patch.object(skool2html, 'write_disassembly', mock_write_disassembly)
+    def test_option_O(self):
+        skoolfile = self.write_text_file(suffix='.skool')
+        for option in ('-O', '--rebuild-audio'):
+            output, error = self.run_skool2html('{} {}'.format(option, skoolfile))
+            self.assertEqual(error, '')
+            self.assertTrue(html_writer.file_info.replace_audio)
+
     @patch.object(skool2html, 'get_object', Mock(return_value=TestHtmlWriter))
     @patch.object(skool2html, 'SkoolParser', MockSkoolParser)
     @patch.object(skool2html, 'write_disassembly', mock_write_disassembly)
@@ -1257,7 +1303,7 @@ class Skool2HtmlTest(SkoolKitTestCase):
         for option in ('-R', '--ref-file'):
             output, error = self.run_skool2html(option, catch_exit=0)
             self.assertEqual(error, '')
-            self.assertTrue(output.startswith('[Colours]\n'))
+            self.assertTrue(output.startswith('[AudioWriter]\n'))
             self.assertIn('\n[Titles]\n', output)
 
     def test_option_r(self):
@@ -1322,6 +1368,14 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertEqual(options.base, 16)
         self.assertEqual(options.case, 2)
 
+    @patch.object(skool2html, 'run', mock_run)
+    @patch.object(skool2html, 'get_config', mock_config)
+    def test_option_I_updates_config_params(self):
+        self.run_skool2html('-I EntryLabel=M_{address} -I EntryPointLabel=EP_{address} test.skool')
+        config = run_args[2]
+        self.assertEqual(config['EntryLabel'], 'M_{address}')
+        self.assertEqual(config['EntryPointLabel'], 'EP_{address}')
+
     @patch.object(skool2html, 'run', mock_run)
     @patch.object(skool2html, 'get_config', mock_config)
     def test_option_I_invalid_value(self):
@@ -1340,9 +1394,12 @@ class Skool2HtmlTest(SkoolKitTestCase):
             Base=0
             Case=0
             CreateLabels=0
+            EntryLabel=L{address}
+            EntryPointLabel={main}_{index}
             JoinCss=
             OutputDir=.
             Quiet=0
+            RebuildAudio=0
             RebuildImages=0
             Search=
             Theme=
@@ -1368,9 +1425,12 @@ class Skool2HtmlTest(SkoolKitTestCase):
             Base=0
             Case=0
             CreateLabels=0
+            EntryLabel=L{address}
+            EntryPointLabel={main}_{index}
             JoinCss=
             OutputDir=html
             Quiet=1
+            RebuildAudio=0
             RebuildImages=0
             Search=
             Theme=dark,wide
@@ -1417,6 +1477,25 @@ class Skool2HtmlTest(SkoolKitTestCase):
         self.assertTrue(error.startswith('usage: skool2html.py'))
         self.assertTrue(error.endswith("missing variable value: 'foo'\n"))
 
+    @patch.object(skool2html, 'write_disassembly', mock_write_disassembly)
+    @patch.object(skool2html, 'get_config', mock_config)
+    def test_Config_InitModule_parameter(self):
+        init_module = """
+            from skoolkit.skoolhtml import HtmlWriter
+            HtmlWriter.version = 1
+        """
+        ref = "[Config]\nInitModule={0}:{1}"
+        module_dir = self.make_directory()
+        module_path = os.path.join(module_dir, 'initmod.py')
+        module = self.write_text_file(dedent(init_module).strip(), path=module_path)
+        module_name = os.path.basename(module)[:-3]
+        reffile = self.write_text_file(ref.format(module_dir, module_name), suffix='.ref')
+        skoolfile = self.write_text_file(path='{}.skool'.format(reffile[:-4]))
+        output, error = self.run_skool2html(skoolfile)
+        self.assertEqual(error, '')
+        html_writer = write_disassembly_args[0]
+        self.assertEqual(getattr(html_writer, "version"), 1)
+
     @patch.object(skool2html, 'get_object', Mock(return_value=TestHtmlWriter))
     @patch.object(skool2html, 'SkoolParser', MockSkoolParser)
     @patch.object(skool2html, 'write_disassembly', mock_write_disassembly)
