@@ -7,7 +7,7 @@ from skoolkit import components, sna2skool, snapshot, SkoolKitError, VERSION
 from skoolkit.config import COMMANDS
 
 def mock_make_snapshot(fname, org, start, end, page):
-    return [0] * 65536, max(16384, start), end
+    return [0] * 65536, 16384 if start is None else start, end
 
 class MockCtlParser:
     def __init__(self, ctls=None):
@@ -30,9 +30,7 @@ class MockSkoolWriter:
         self.config = config
         self.wrote_skool = False
 
-    def write_skool(self, write_refs, text):
-        self.write_refs = write_refs
-        self.text = text
+    def write_skool(self):
         self.wrote_skool = True
 
 def mock_run(*args):
@@ -43,6 +41,14 @@ def mock_config(name):
     return {k: v[0] for k, v in COMMANDS[name].items()}
 
 class Sna2SkoolTest(SkoolKitTestCase):
+    def setUp(self):
+        super().setUp()
+        self.cwd = os.getcwd()
+
+    def tearDown(self):
+        os.chdir(self.cwd)
+        super().tearDown()
+
     @patch.object(sna2skool, 'run', mock_run)
     @patch.object(sna2skool, 'get_config', mock_config)
     def test_default_option_values(self):
@@ -50,10 +56,11 @@ class Sna2SkoolTest(SkoolKitTestCase):
         sna2skool.main((sna,))
         snafile, options = run_args[:2]
         self.assertEqual(snafile, sna)
-        self.assertEqual([], options.ctlfiles)
+        self.assertEqual([], options.ctls)
+        self.assertEqual(None, options.defb)
         self.assertEqual(options.base, 10)
         self.assertEqual(options.case, 2)
-        self.assertEqual(options.start, 0)
+        self.assertIsNone(options.start)
         self.assertEqual(options.end, 65536)
         self.assertIsNone(options.org)
         self.assertIsNone(options.page)
@@ -71,6 +78,7 @@ class Sna2SkoolTest(SkoolKitTestCase):
             LineWidth=119
             ListRefs=2
             Text=1
+            Timings=1
             Title-b=Data at {address}
             Title-c=Code at {address}
         """
@@ -79,10 +87,10 @@ class Sna2SkoolTest(SkoolKitTestCase):
         sna2skool.main((sna,))
         snafile, options, config = run_args
         self.assertEqual(snafile, sna)
-        self.assertEqual([], options.ctlfiles)
+        self.assertEqual([], options.ctls)
         self.assertEqual(options.base, 16)
         self.assertEqual(options.case, 1)
-        self.assertEqual(options.start, 0)
+        self.assertIsNone(options.start)
         self.assertEqual(options.end, 65536)
         self.assertIsNone(options.org)
         self.assertIsNone(options.page)
@@ -91,6 +99,7 @@ class Sna2SkoolTest(SkoolKitTestCase):
         self.assertEqual(config.get('DefbSize'), 12)
         self.assertEqual(options.line_width, 119)
         self.assertEqual(config.get('DefmSize'), 92)
+        self.assertEqual(config.get('Timings'), 1)
         self.assertEqual(config.get('Title-b'), 'Data at {address}')
         self.assertEqual(config.get('Title-c'), 'Code at {address}')
 
@@ -219,6 +228,99 @@ class Sna2SkoolTest(SkoolKitTestCase):
         self.assertEqual(mock_ctl_parser.ctls, {65533: 'c', 65536: 'i'})
         self.assertTrue(mock_skool_writer.wrote_skool)
 
+    @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_c_with_directory(self):
+        ctldir = self.make_directory()
+        ctlfiles = [self.write_text_file(path=os.path.join(ctldir, c)) for c in ('bar.ctl', 'foo.ctl')]
+        self.write_text_file(path='{}/not-a-ctl.txt'.format(ctldir))
+        output, error = self.run_sna2skool('-c {} test-c-dir.bin'.format(ctldir))
+        self.assertEqual(error, 'Using control files: {}\n'.format(', '.join(ctlfiles)))
+        self.assertEqual(ctlfiles, mock_ctl_parser.ctlfiles)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_ctl_with_two_directories(self):
+        ctldir1 = self.make_directory()
+        ctlfiles = [self.write_text_file(path=os.path.join(ctldir1, c)) for c in ('bar.ctl', 'foo.ctl')]
+        ctldir2 = self.make_directory()
+        ctlfiles.append(self.write_text_file(path=os.path.join(ctldir2, 'baz.ctl')))
+        output, error = self.run_sna2skool('--ctl {} --ctl {} test-c-dirs.bin'.format(ctldir1, ctldir2))
+        self.assertEqual(error, 'Using control files: {}\n'.format(', '.join(ctlfiles)))
+        self.assertEqual(ctlfiles, mock_ctl_parser.ctlfiles)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_c_with_directory_and_file(self):
+        ctldir = self.make_directory()
+        ctlfiles = [self.write_text_file(path=os.path.join(ctldir, c)) for c in ('bar.ctl', 'foo.ctl')]
+        ctl3 = self.write_text_file(suffix='.ctl')
+        ctlfiles.append(ctl3)
+        output, error = self.run_sna2skool('-c {} -c {} test-c-dir-file.bin'.format(ctldir, ctl3))
+        self.assertEqual(error, 'Using control files: {}\n'.format(', '.join(ctlfiles)))
+        self.assertEqual(ctlfiles, mock_ctl_parser.ctlfiles)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_ctl_with_empty_directory(self):
+        binfile = self.write_bin_file([0] * 2, suffix='.bin')
+        output, error = self.run_sna2skool('--ctl {} {}'.format(self.make_directory(), binfile))
+        self.assertEqual(error, '')
+        self.assertEqual(mock_ctl_parser.ctls, {65534: 'c', 65536: 'i'})
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_c_with_current_directory(self):
+        os.chdir(self.make_directory())
+        ctl1 = self.write_text_file(suffix='.ctl')
+        ctl2 = self.write_text_file(suffix='.ctl')
+        ctlfiles = sorted(os.path.join('.', f) for f in (ctl1, ctl2))
+        output, error = self.run_sna2skool('-c . test-c-cwd.bin')
+        self.assertEqual(error, 'Using control files: {}\n'.format(', '.join(ctlfiles)))
+        self.assertEqual(ctlfiles, mock_ctl_parser.ctlfiles)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_d(self):
+        for option, value in (('-d', 4), ('--defb', 16)):
+            output, error = self.run_sna2skool(f'{option} {value} -s 32768 -e 49152 test-defb.bin')
+            self.assertEqual(error, '')
+            self.assertEqual(mock_ctl_parser.ctls, {32768: 'b', 49152: 'i'})
+            self.assertEqual(mock_skool_writer.config['DefbSize'], value)
+            self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_d_overrides_default_ctl_file(self):
+        binfile = self.write_bin_file([0], suffix='.bin')
+        self.write_text_file('c 65535', f'{binfile[:-4]}.ctl')
+        output, error = self.run_sna2skool(f'-d 1 {binfile}')
+        self.assertEqual(error, '')
+        self.assertEqual(mock_ctl_parser.ctls, {65535: 'b', 65536: 'i'})
+        self.assertEqual(mock_skool_writer.config['DefbSize'], 1)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_defb_overrides_option_ctl(self):
+        binfile = self.write_bin_file([0], suffix='.bin')
+        ctlfile = self.write_text_file('c 65535', suffix='.ctl')
+        output, error = self.run_sna2skool(f'--defb 1 --ctl {ctlfile} {binfile}')
+        self.assertEqual(error, '')
+        self.assertEqual(mock_ctl_parser.ctls, {65535: 'b', 65536: 'i'})
+        self.assertEqual(mock_skool_writer.config['DefbSize'], 1)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
     @patch.object(sna2skool, 'CtlParser', MockCtlParser)
     @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
     def test_option_e(self):
@@ -352,7 +454,7 @@ class Sna2SkoolTest(SkoolKitTestCase):
             Case=2
             CommentWidthMin=10
             DefbSize=8
-            DefmSize=66
+            DefmSize=65
             DefwSize=1
             EntryPointRef=This entry point is used by the routine at {ref}.
             EntryPointRefs=This entry point is used by the routines at {refs} and {ref}.
@@ -360,9 +462,11 @@ class Sna2SkoolTest(SkoolKitTestCase):
             LineWidth=79
             ListRefs=1
             Ref=Used by the routine at {ref}.
+            RefFormat=#R{address}
             Refs=Used by the routines at {refs} and {ref}.
             Semicolons=c
             Text=0
+            Timings=0
             Title-b=Data block at {address}
             Title-c=Routine at {address}
             Title-g=Game status buffer entry at {address}
@@ -371,6 +475,7 @@ class Sna2SkoolTest(SkoolKitTestCase):
             Title-t=Message at {address}
             Title-u=Unused
             Title-w=Data block at {address}
+            Wrap=0
         """
         self.assertEqual(dedent(exp_output).strip(), output.rstrip())
 
@@ -390,7 +495,7 @@ class Sna2SkoolTest(SkoolKitTestCase):
             Case=1
             CommentWidthMin=10
             DefbSize=8
-            DefmSize=66
+            DefmSize=65
             DefwSize=1
             EntryPointRef=This entry point is used by the routine at {ref}.
             EntryPointRefs=This entry point is used by the routines at {refs} and {ref}.
@@ -398,9 +503,11 @@ class Sna2SkoolTest(SkoolKitTestCase):
             LineWidth=79
             ListRefs=1
             Ref=Called by the routine at {ref}.
+            RefFormat=#R{address}
             Refs=Used by the routines at {refs} and {ref}.
             Semicolons=c
             Text=0
+            Timings=0
             Title-b=Data block at {address}
             Title-c=Routine at {address}
             Title-g=Game status buffer entry at {address}
@@ -409,9 +516,32 @@ class Sna2SkoolTest(SkoolKitTestCase):
             Title-t=Text at {address}
             Title-u=Unused
             Title-w=Data block at {address}
+            Wrap=0
         """
         self.assertEqual(dedent(exp_output).strip(), output.rstrip())
 
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_start_address_defaults_to_16384_for_snapshot(self):
+        snafile = self.write_bin_file([0] * 49179, suffix='.sna')
+        output, error = self.run_sna2skool('{}'.format(snafile))
+        self.assertEqual(error, '')
+        exp_ctls = {16384: 'c', 65536: 'i'}
+        self.assertEqual(exp_ctls, mock_ctl_parser.ctls)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_start_address_defaults_to_org_address_for_raw_memory_file(self):
+        data = [0] * 10
+        binfile = self.write_bin_file(data, suffix='.bin')
+        org = 37
+        output, error = self.run_sna2skool('-o {} {}'.format(org, binfile))
+        self.assertEqual(error, '')
+        exp_ctls = {org: 'c', org + len(data): 'i'}
+        self.assertEqual(exp_ctls, mock_ctl_parser.ctls)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
     @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
     @patch.object(sna2skool, 'CtlParser', MockCtlParser)
     @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
@@ -436,6 +566,17 @@ class Sna2SkoolTest(SkoolKitTestCase):
             self.assertEqual(exp_ctls, mock_ctl_parser.ctls)
             self.assertTrue(mock_skool_writer.wrote_skool)
 
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_option_s_below_16384_with_snapshot(self):
+        start = 8192
+        exp_ctls = {start: 'c', 65536: 'i'}
+        snafile = self.write_bin_file([0] * 49179, suffix='.sna')
+        output, error = self.run_sna2skool('-s {} {}'.format(start, snafile))
+        self.assertEqual(error, '')
+        self.assertEqual(exp_ctls, mock_ctl_parser.ctls)
+        self.assertTrue(mock_skool_writer.wrote_skool)
+
     @patch.object(sna2skool, 'make_snapshot', mock_make_snapshot)
     @patch.object(sna2skool, 'CtlParser', MockCtlParser)
     @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
@@ -535,3 +676,16 @@ class Sna2SkoolTest(SkoolKitTestCase):
         self.assertEqual(error, '')
         self.assertEqual([128], mock_skool_writer.snapshot)
         self.assertTrue(mock_skool_writer.wrote_skool)
+
+    @patch.object(components, 'SK_CONFIG', None)
+    @patch.object(sna2skool, 'CtlParser', MockCtlParser)
+    @patch.object(sna2skool, 'SkoolWriter', MockSkoolWriter)
+    def test_custom_default_disassembly_start_address(self):
+        ini = "[skoolkit]\nDefaultDisassemblyStartAddress=24576"
+        self.write_text_file(ini, 'skoolkit.ini')
+        snafile = self.write_bin_file([0] * 49179, suffix='.sna')
+        output, error = self.run_sna2skool(snafile)
+        self.assertEqual(error, '')
+        exp_ctls = {24576: 'c', 65536: 'i'}
+        self.assertEqual(exp_ctls, mock_ctl_parser.ctls)
+        self.assertTrue(mock_skool_writer.wrote_skool)
