@@ -39,6 +39,29 @@ class Tap2SnaTest(SkoolKitTestCase):
             tzx_data.extend(block)
         return self.write_bin_file(tzx_data, suffix='.tzx')
 
+    def _write_basic_loader(self, start, data, write=True, program='simloadbas', code='simloadbyt'):
+        start_str = [ord(c) for c in str(start)]
+        basic_data = [
+            0, 10,            # Line 10
+            16, 0,            # Line length
+            239, 34, 34, 175, # LOAD ""CODE
+            58,               # :
+            249, 192, 176,    # RANDOMIZE USR VAL
+            34,               # "
+            *start_str,       # start address
+            34,               # "
+            13                # ENTER
+        ]
+        blocks = [
+            create_tap_header_block(program, 10, len(basic_data), 0),
+            create_tap_data_block(basic_data),
+            create_tap_header_block(code, start, len(data)),
+            create_tap_data_block(data)
+        ]
+        if write:
+            return self._write_tap(blocks), basic_data
+        return blocks, basic_data
+
     def _get_snapshot(self, start=16384, data=None, options='', load_options=None, blocks=None, tzx=False):
         if blocks is None:
             blocks = [create_tap_data_block(data)]
@@ -299,6 +322,31 @@ class Tap2SnaTest(SkoolKitTestCase):
         self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
         self.assertEqual(code, snapshot[code_start:code_start + len(code)])
 
+    def test_empty_standard_speed_data_block_in_tzx_file_is_ignored(self):
+        basic_data = [6, 7]
+        code_start = 49152
+        code = [8, 9, 10]
+        empty_block = [
+            16,   # Standard speed data
+            0, 0, # Pause (0ms)
+            0, 0, # Data length (0)
+        ]
+        blocks = [
+            create_tzx_header_block(data_type=0),
+            create_tzx_data_block(basic_data),
+            empty_block,
+            create_tzx_header_block(start=code_start),
+            create_tzx_data_block(code)
+        ]
+
+        tzxfile = self._write_tzx(blocks)
+        z80file = self.write_bin_file(suffix='.z80')
+        output, error = self.run_tap2sna('--force {} {}'.format(tzxfile, z80file))
+        self.assertEqual(error, '')
+        snapshot = get_snapshot(z80file)
+        self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
+        self.assertEqual(code, snapshot[code_start:code_start + len(code)])
+
     def test_ram_call(self):
         ram_module = """
             def fix(snapshot):
@@ -712,26 +760,8 @@ class Tap2SnaTest(SkoolKitTestCase):
     @patch.object(tap2sna, '_write_z80', mock_write_z80)
     def test_sim_load(self):
         code_start = 32768
-        code_start_str = [ord(c) for c in str(code_start)]
-        basic_data = [
-            0, 10,            # Line 10
-            16, 0,            # Line length
-            239, 34, 34, 175, # LOAD ""CODE
-            58,               # :
-            249, 192, 176,    # RANDOMIZE USR VAL
-            34,               # "
-            *code_start_str,  # start address
-            34,               # "
-            13                # ENTER
-        ]
         code = [4, 5]
-        blocks = [
-            create_tap_header_block("simloadbas", 10, len(basic_data), 0),
-            create_tap_data_block(basic_data),
-            create_tap_header_block("simloadbyt", code_start, len(code)),
-            create_tap_data_block(code)
-        ]
-        tapfile = self._write_tap(blocks)
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
         z80file = '{}/out.z80'.format(self.make_directory())
         output, error = self.run_tap2sna(f'--sim-load {tapfile} {z80file}')
         out_lines = output.strip().split('\n')
@@ -781,26 +811,8 @@ class Tap2SnaTest(SkoolKitTestCase):
     def test_sim_load_with_given_start_address(self):
         code_start = 32768
         start = 32769
-        code_start_str = [ord(c) for c in str(code_start)]
-        basic_data = [
-            0, 10,            # Line 10
-            16, 0,            # Line length
-            239, 34, 34, 175, # LOAD ""CODE
-            58,               # :
-            249, 192, 176,    # RANDOMIZE USR VAL
-            34,               # "
-            *code_start_str,  # start address
-            34,               # "
-            13                # ENTER
-        ]
         code = [175, 201]
-        blocks = [
-            create_tap_header_block("simloadbas", 10, len(basic_data), 0),
-            create_tap_data_block(basic_data),
-            create_tap_header_block("simloadbyt", code_start, len(code)),
-            create_tap_data_block(code)
-        ]
-        tapfile = self._write_tap(blocks)
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
         z80file = '{}/out.z80'.format(self.make_directory())
         output, error = self.run_tap2sna(f'--sim-load --start {start} {tapfile} {z80file}')
         out_lines = output.strip().split('\n')
@@ -1162,6 +1174,32 @@ class Tap2SnaTest(SkoolKitTestCase):
         exp_reg = set(('SP=65344', 'IX=49154', 'IY=23610', 'PC=32780', 'F=1'))
         self.assertLessEqual(exp_reg, set(options.reg))
 
+    @patch.object(tap2sna, '_write_z80', mock_write_z80)
+    def test_sim_load_ignores_extra_byte_at_end_of_tape(self):
+        code_start = 32768
+        code = [4, 5]
+        blocks, basic_data = self._write_basic_loader(code_start, code, False)
+        tapfile = self._write_tap(blocks + [[0]])
+        z80file = '{}/out.z80'.format(self.make_directory())
+        output, error = self.run_tap2sna(f'--sim-load {tapfile} {z80file}')
+        out_lines = output.strip().split('\n')
+        exp_out_lines = [
+            'Program: simloadbas',
+            'Fast loading data block: 23755,20',
+            '',
+            'Bytes: simloadbyt',
+            'Fast loading data block: 32768,2',
+            '',
+            'Tape finished',
+            'Simulation stopped (PC in RAM): PC=32768',
+        ]
+        self.assertEqual(exp_out_lines, out_lines)
+        self.assertEqual(error, '')
+        self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
+        self.assertEqual(code, snapshot[code_start:code_start + len(code)])
+        exp_reg = set(('SP=65344', 'IX=32770', 'IY=23610', 'PC=32768'))
+        self.assertLessEqual(exp_reg, set(options.reg))
+
     @patch.object(tap2sna, '_write_z80', mock_write_z80)
     def test_sim_load_preserves_border_colour(self):
         code_start = 32768
@@ -1208,33 +1246,14 @@ class Tap2SnaTest(SkoolKitTestCase):
     @patch.object(tap2sna, '_write_z80', mock_write_z80)
     def test_sim_load_preserves_interrupt_mode_and_flip_flop(self):
         code_start = 32768
-        code_start_str = [ord(c) for c in str(code_start)]
-        basic_data = [
-            0, 10,            # Line 10
-            16, 0,            # Line length
-            239, 34, 34, 175, # LOAD ""CODE
-            58,               # :
-            249, 192, 176,    # RANDOMIZE USR VAL
-            34,               # "
-            *code_start_str,  # start address
-            34,               # "
-            13                # ENTER
-        ]
         code = [
             243,              # 32768 DI
             237, 94,          # 32769 IM 2
             201,              # 32771 RET
         ]
         start = 32771
-        blocks = [
-            create_tap_header_block("simloadbas", 10, len(basic_data), 0),
-            create_tap_data_block(basic_data),
-            create_tap_header_block("simloadbyt", code_start, len(code)),
-            create_tap_data_block(code)
-        ]
-        tapfile = self._write_tap(blocks)
-        # z80file = '{}/out.z80'.format(self.make_directory())
-        z80file = 'out.z80'
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
+        z80file = '{}/out.z80'.format(self.make_directory())
         output, error = self.run_tap2sna(f'--sim-load --start {start} {tapfile} {z80file}')
         out_lines = output.strip().split('\n')
         exp_out_lines = [
@@ -1252,6 +1271,141 @@ class Tap2SnaTest(SkoolKitTestCase):
         self.assertIn('im=2', options.state)
         self.assertIn('iff=0', options.state)
 
+    @patch.object(tap2sna, '_write_z80', mock_write_z80)
+    def test_sim_load_with_ram_call(self):
+        ram_module = """
+            def fix(snapshot):
+                snapshot[32768:32772] = [1, 2, 3, 4]
+        """
+        module_dir = self.make_directory()
+        module_name = 'simloadram'
+        module_path = os.path.join(module_dir, f'{module_name}.py')
+        module = self.write_text_file(textwrap.dedent(ram_module).strip(), path=module_path)
+        code_start = 32768
+        code = [4, 5]
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
+        output, error = self.run_tap2sna(f'--sim-load --ram call={module_dir}:{module_name}.fix {tapfile} out.z80')
+        out_lines = output.strip().split('\n')
+        exp_out_lines = [
+            'Program: simloadbas',
+            'Fast loading data block: 23755,20',
+            '',
+            'Bytes: simloadbyt',
+            'Fast loading data block: 32768,2',
+            '',
+            'Tape finished',
+            'Simulation stopped (PC in RAM): PC=32768',
+        ]
+        self.assertEqual(exp_out_lines, out_lines)
+        self.assertEqual(error, '')
+        self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
+        self.assertEqual([1, 2, 3, 4], snapshot[32768:32772])
+        exp_reg = set(('SP=65344', 'IX=32770', 'IY=23610', 'PC=32768'))
+        self.assertLessEqual(exp_reg, set(options.reg))
+
+    @patch.object(tap2sna, '_write_z80', mock_write_z80)
+    def test_sim_load_with_ram_move(self):
+        code_start = 32768
+        code = [4, 5]
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
+        output, error = self.run_tap2sna(f'--sim-load --ram move=32768,2,32770 {tapfile} out.z80')
+        out_lines = output.strip().split('\n')
+        exp_out_lines = [
+            'Program: simloadbas',
+            'Fast loading data block: 23755,20',
+            '',
+            'Bytes: simloadbyt',
+            'Fast loading data block: 32768,2',
+            '',
+            'Tape finished',
+            'Simulation stopped (PC in RAM): PC=32768',
+        ]
+        self.assertEqual(exp_out_lines, out_lines)
+        self.assertEqual(error, '')
+        self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
+        self.assertEqual([4, 5, 4, 5], snapshot[32768:32772])
+        exp_reg = set(('SP=65344', 'IX=32770', 'IY=23610', 'PC=32768'))
+        self.assertLessEqual(exp_reg, set(options.reg))
+
+    @patch.object(tap2sna, '_write_z80', mock_write_z80)
+    def test_sim_load_with_ram_poke(self):
+        code_start = 32768
+        code = [4, 5]
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
+        output, error = self.run_tap2sna(f'--sim-load --ram poke=32768-32770-2,1 {tapfile} out.z80')
+        out_lines = output.strip().split('\n')
+        exp_out_lines = [
+            'Program: simloadbas',
+            'Fast loading data block: 23755,20',
+            '',
+            'Bytes: simloadbyt',
+            'Fast loading data block: 32768,2',
+            '',
+            'Tape finished',
+            'Simulation stopped (PC in RAM): PC=32768',
+        ]
+        self.assertEqual(exp_out_lines, out_lines)
+        self.assertEqual(error, '')
+        self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
+        self.assertEqual([1, 5, 1], snapshot[32768:32771])
+        exp_reg = set(('SP=65344', 'IX=32770', 'IY=23610', 'PC=32768'))
+        self.assertLessEqual(exp_reg, set(options.reg))
+
+    @patch.object(tap2sna, '_write_z80', mock_write_z80)
+    def test_sim_load_with_ram_sysvars(self):
+        code_start = 32768
+        code = [4, 5]
+        tapfile, basic_data = self._write_basic_loader(code_start, code)
+        output, error = self.run_tap2sna(f'--sim-load --ram sysvars {tapfile} out.z80')
+        out_lines = output.strip().split('\n')
+        exp_out_lines = [
+            'Program: simloadbas',
+            'Fast loading data block: 23755,20',
+            '',
+            'Bytes: simloadbyt',
+            'Fast loading data block: 32768,2',
+            '',
+            'Tape finished',
+            'Simulation stopped (PC in RAM): PC=32768',
+        ]
+        self.assertEqual(exp_out_lines, out_lines)
+        self.assertEqual(error, '')
+        self.assertEqual(basic_data, snapshot[23755:23755 + len(basic_data)])
+        self.assertEqual([203, 92], snapshot[23627:23629]) # VARS=23755
+        self.assertEqual([204, 92], snapshot[23641:23643]) # E-LINE=23756
+        self.assertEqual([206, 92], snapshot[23649:23651]) # WORKSP=23758
+        self.assertEqual([206, 92], snapshot[23651:23653]) # STKBOT=23758
+        self.assertEqual([206, 92], snapshot[23653:23655]) # STKEND=23758
+        exp_reg = set(('SP=65344', 'IX=32770', 'IY=23610', 'PC=32768'))
+        self.assertLessEqual(exp_reg, set(options.reg))
+
+    @patch.object(tap2sna, '_write_z80', mock_write_z80)
+    def test_sim_load_with_unexpected_end_of_tape(self):
+        basic_data = [
+            0, 10,       # Line 10
+            4, 0,        # Line length
+            239, 34, 34, # LOAD ""
+            13           # ENTER
+        ]
+        blocks = [
+            create_tap_header_block("simloadbas", 10, len(basic_data), 0),
+            create_tap_data_block(basic_data),
+        ]
+        tapfile = self._write_tap(blocks)
+        z80file = '{}/out.z80'.format(self.make_directory())
+        with self.assertRaises(SkoolKitError) as cm:
+            self.run_tap2sna(f'--sim-load {tapfile} {z80file}')
+        out_lines = self.out.getvalue().strip().split('\n')
+        exp_out_lines = [
+            'Program: simloadbas',
+            'Fast loading data block: 23755,8',
+            '',
+            'Tape finished'
+        ]
+        self.assertEqual(exp_out_lines, out_lines)
+        self.assertEqual(cm.exception.args[0], 'Error while getting snapshot out.z80: Failed to fast load block: unexpected end of tape')
+        self.assertEqual(self.err.getvalue(), '')
+
     @patch.object(tap2sna, '_write_z80', mock_write_z80)
     def test_sim_load_with_tzx_block_type_0x15(self):
         block = [
